'''
Module to handle distortions in diffraction patters.
'''

import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize


def filter_ring(points, center, rminmax):
    '''Filter points to be in a certain radial distance range from center.
    
    Parameters:
        points (np.ndarray):    Candidate points.
        center (np.ndarray/tuple):    Center position.
        rminmax (tuple):    Tuple of min and max radial distance.
    
    Returns:
        (np.ndarray):    List of filtered points, two column array.
        
    '''
    
    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
        
        rminmax = np.array(rminmax)
        rminmax = np.reshape(rminmax, 2)
   
    except:
        raise TypeError('Something wrong with the input!')
    
    # calculate radii
    rs = np.sqrt( np.square(points[:,0]-center[0]) + np.square(points[:,1]-center[1]) )
    
    # filter by given limits
    sel = (rs>=rminmax[0])*(rs<=rminmax[1])
    
    if sel.any():
        return points[sel]
    else:
        return None
    
    
def points_topolar(points, center):
    '''Convert points to polar coordinate system.
    
    Can be either in pixel or real dim, but should be the same for points and center.
    
    Parameters:
        points (np.ndarray):    Positions as two column array.
        center (np.ndarray/tuple):    Origin of the polar coordinate system.
    
    Returns:
        (np.ndarray):    Positions in polar coordinate system as two column array (r, theta).
    
    '''
    
    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
   
    except:
        raise TypeError('Something wrong with the input!')
    
    # calculate radii
    rs = np.sqrt( np.square(points[:,0]-center[0]) + np.square(points[:,1]-center[1]) )
    # calculate angle
    thes = np.arctan2(points[:,1]-center[1], points[:,0]-center[0])
    
    return np.array( [rs, thes] ).transpose()
    
    
def plot_ringpolar(points, dims, show=False):
    '''Plot points in polar coordinate system.
    
    Parameters:
        points (np.ndarrad):    Positions in polar coords.
        dims (tuple):    Dimension information to plot labels.
        show (bool):    Set to directly show plot in interactive mode.
       
    Returns:
        (np.ndarray):    Image of the plot.   
        
    '''

    try:
        # try to convert input to np.ndarray with 2 columns (necessary if only one entry provided)
        points = np.reshape(np.array(points), (-1,2))
        # check if enough dims availabel
        assert(len(dims)>=2)
        assert(len(dims[0])==3)
    except:
        raise TypeError('Something wrong with the input!')
        
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    # mean value as line
    ax.axhline(np.mean(points[:,0]), ls='--', c='k')
    
    # points
    ax.plot(points[:,1], points[:,0], 'rx')
    
    # labels
    ax.set_xlabel('theta /[rad]')
    ax.set_xlim( (-np.pi, np.pi) )
    ax.set_ylabel('r /{}'.format(dims[0][2]))
    
    if show:
        plt.show(block=False)
    
    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    return plot


def residuals_center( param, data):
    '''Residual function for minimizing the deviations from the mean radial distance.
    
    Parameters:
        param (np.ndarray):    The center to optimize.
        data (np.ndarray):    The points in x,y coordinates of the original image.
        
    Returns:
        (np.ndarray):   Residuals.    
        
    '''
    
    # manually calculating the radii, as we do not need the thetas
    rs = np.sqrt( np.square(data[:,0]-param[0]) + np.square(data[:,1]-param[1]) )
    
    return (rs-np.mean(rs))
    
    
def optimize_center(points, center, maxfev=1000, verbose=None):
    '''Optimize the center by minimizing the sum of square deviations from the mean radial distance.
    
    Parameters:
        points (np.ndarray):    The points to which the optimization is done (x,y coords in org image).
        center (np.ndarray/tuple):    Initial center guess.
        maxfev (int):    Max number of iterations forwarded to scipy.optimize.leastsq().
        verbose (bool):    Set to get verbose output.
    
    Returns:
        (np.ndarray):    The optimized center.
    
    '''

    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
   
    except:
        raise TypeError('Something wrong with the input!')

    # run the optimization
    popt, flag = scipy.optimize.leastsq( residuals_center, center, args=(points), maxfev=maxfev)
    
    if flag not in [1,2,3,4]:
        print('WARNING: center optimization failed.')

    if verbose:
        print('optimized center: ({}, {})'.format(center[0], center[1]))

    return popt
    

def rad_dis( theta, alpha, beta, order=2 ):
    '''Radial distortion due to ellipticity or higher order distortion.
    
    Relative distortion, to be multiplied with radial distance.
    
    Parameters:
        theta (np.ndarray/float):    Angles at which to evaluate.
        alpha (float):    Orientation of major axis.
        beta (float):    Strength of distortion (beta = (1-r_min/r_max)/(1+r_min/r_max).
        order (int):    Order of distortion.
        
    Returns:
        (np.ndarray/float):    Distortion factor.
        
    '''
    
    return (1.-np.square(beta))/np.sqrt(1.+np.square(beta)-2.*beta*np.cos(order*(theta+alpha)))

    
def residuals_dis(param, points, ns):
    '''Residual function for distortions.
    
    Parameters:
        param (np.ndarray):    Parameters for distortion.
        points (np.ndarray):    Points to fit to.
        ns (tuple):    List of orders to account for.
    
    Returns:
        (np.ndarray):   Residuals.
        
    '''

    est = param[0]*np.ones(points[:,1].shape)
    for i in range(len(ns)):
        est *=rad_dis( points[:,1], param[i*2+1], param[i*2+2], ns[i])
            
    return points[:,0] - est 
    
    
def optimize_distortion(points, ns, maxfev=1000, verbose=False):
    '''Optimize distortions.
    
    The orders in the list ns are first fitted subsequently and the result is refined in a final fit simultaneously fitting all orders.
    
    Parameters:
        points (np.ndarray):    Points to optimize to (in polar coords).
        ns (tuple):    List of orders to correct for.
        maxfev (int):    Max number of iterations forwarded to scipy.optimize.leastsq().
        verbose (bool):    Set for verbose output.
    
    Returns:
        (np.ndarray):    Optimized parameters according to ns.
        
    '''
    
    try:
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        # check points to be sufficient for fitting
        assert(points.shape[0] >= 3)
        
        # check orders
        assert(len(ns)>=1)
    except:
        raise TypeError('Something wrong with the input!')
        
    
    # init guess for full fit
    init_guess = np.ones(len(ns)*2+1)
    init_guess[0] = np.mean(points[:,0])
    
    # make a temporary copy
    points_tmp = np.copy(points)
    
    if verbose:
        print('correction for {} order distortions.'.format(ns))
        print('starting with subsequent fitting:')
    
    # subsequently fit the orders
    for i in range(len(ns)):
        # optimize order to points_tmp
        popt, flag = scipy.optimize.leastsq( residuals_dis, (init_guess[0], 0.1, 0.1), args=(points_tmp, (ns[i],)), maxfev=maxfev)
        
        if flag not in [1,2,3,4]:
            print('WARNING: optimization of distortions failed.')
        
        # information
        if verbose:
            print('fitted order {}: R={} alpha={} beta={}'.format(ns[i], popt[0], popt[1], popt[2]))
        
        # save for full fit
        init_guess[i*2+1] = popt[1]
        init_guess[i*2+2] = popt[2]
        
        # do correction
        points_tmp[:,0] /= rad_dis(points_tmp[:,1], popt[1], popt[2], ns[i])
    
    # full fit    
    if verbose:
        print('starting the full fit:')    
    
    popt, flag = scipy.optimize.leastsq( residuals_dis, init_guess, args=(points, ns), maxfev=maxfev)
    
    if flag not in [1,2,3,4]:
        print('WARNING: optimization of distortions failed.')
    
    if verbose:
        print('fitted to: R={}'.format(popt[0]))
        for i in range(len(ns)):
            print('.. order={}, alpha={}, beta={}'.format(ns[i], popt[i*2+1], popt[i*2+2]))

    return popt
    
    
def plot_distpolar(points, dims, dists, ns, show=False):
    '''Plot the results of distortion fitting in polar coordinates.
    
    Parameters:
        points (np.ndarray):    Points in polar coords.
        dims (tuple):    Dimensions, necessary to have unit information.
        dists (np.ndarray):    Results of dist fitting, length according to ns.
        ns (list):    List of used orders.
        show (bool):    Set to directly show the plot in interactive mode.
        
    Returns:
        (np.ndarray):    Image of the plot.
        
    '''
    
    try:
        # check points
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        
        # check if enough dims availabel
        assert(len(dims)>=2)
        assert(len(dims[0])==3)
        
        # check orders
        assert(len(ns)>=1)
        
        # check dists
        assert(dists.shape[0] == len(ns)*2+1)
    except:
        raise TypeError('Something wrong with the input!')
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    # stuff from the single orders
    ax.axhline(dists[0], ls='--', c='k')
    xpl_ell = np.linspace(-np.pi, np.pi, 100)
    for i in range(len(ns)):
        plt.plot( xpl_ell, dists[0]*rad_dis(xpl_ell, dists[i*2+1], dists[i*2+2], ns[i]), 'm--')

    # points before
    ax.plot(points[:,1], points[:,0], 'rx')
    
    # sum of all distorts
    sum_dists = np.ones(xpl_ell.shape)*dists[0]
    for i in range(len(ns)):
        sum_dists *= rad_dis(xpl_ell, dists[i*2+1], dists[i*2+2], ns[i])
    plt.plot( xpl_ell, sum_dists, 'b-' )
    
    # points after
    points_corr = np.copy(points)
    for i in range(len(ns)):
        points_corr[:,0] /= rad_dis(points[:,1], dists[i*2+1], dists[i*2+2], ns[i])
    plt.plot( points_corr[:,1], points_corr[:,0], 'gx')
    
    # labels
    ax.set_xlabel('theta /[rad]')
    ax.set_xlim( (-np.pi, np.pi) )
    ax.set_ylabel('r /{}'.format(dims[0][2]))
    
    if show:
        plt.show(block=False)
    
    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    return plot




"""
A module to load data and meta data from DM3 and DM4 files into python 
as written by Gatan's Digital Micrograph program.

On Memory mode:
  The fileDM support and "on memory" mode that preoliads the file data in memory
and data read operations during data parsing are performed against memory. This
is critical to matain good performance when the file resides in a parallel file
system (PFS) because latency of seek operations PFSs is very high.
"""

import mmap
from os import stat as fileStats
import os
from os.path import basename as osBasename

import numpy as np


class fileDM:
    '''Init opening the file and reading in the header.
        
    Parameters:
        filename (str): string pointing to the filesystem location of the file.
        
        verbose (bool): if True, debug information is printed.
        
        on_memory (bool): if True, file data is pre-loaded in memory and all data
        parsing is performed against memory. Use this mode if the file
        is in a network based or paralle file system.
    
    Example:
        Simple example for reading data from a single image into memory:
            
            >>> import matplotlib.pyplot as plt
            >>> from ncempy.io import dm
            >>> with dm.fileDM('filename.dm4') as dmFile1:
                dataSet = dmFile1.getDataset(0)
            >>> plt.imshow(dataSet['data'])
        
        Example of multi-image dm3 file:
            
            >>> with dm.fileDM('imageSeries.dm3')as dmFile2:
                series = dmFile2.getDataset(0)
            >>> plt.imshow(series['data'][0,:,:]) #show the first image in the series
    '''
    def __init__(self, filename, verbose = False, on_memory=False):


        self.filename = filename

        # necessary declarations, if something fails
        self.fid = None

        self._on_memory = on_memory

        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string')

        #Add a top level variable to indicate verbosee output for debugging
        self.v = verbose

        # try opening the file
        try:
            if not self._on_memory:
                self.fid = open(filename, 'rb')
            if self._on_memory:
                self._buffer_offset = 0
                # Pre-load the file as a memory map that supports operations
                # similar to a file.
                with open(filename, 'rb') as _fid:
                    if os.name == 'nt':
                        self.fid=mmap.mmap(_fid.fileno(), 0,
                                           access=mmap.ACCESS_READ)
                    else:
                        self.fid=mmap.mmap(_fid.fileno(), 0,
                                           prot=mmap.PROT_READ) #, flags=mmap.MAP_PRIVATE)
                    self._buffer_size=fileStats(filename).st_size

        except IOError:
            print('Error reading file: "{}"'.format(filename))
            raise
        except :
            raise

        if not self._validDM():
            #print('Not a valid DM3 or DM4 file: "{}"'.format(filename))
            raise IOError('Can not read file: {}'.format(filename))

        #Lists that will contain information about binary data arrays
        self.xSize = []
        self.ySize = []
        self.zSize = []
        self.zSize2 = [] #only used for 4D datasets in DM4 files
        self.dataType = []
        self.dataSize = []
        self.dataOffset = []
        self.dataShape = [] #1,2,3, or 4. The total number of dimensions in a data set

        #The number of objects found in the DM3 file
        self.numObjects = 0

        #Indicator that a thumbnail exists (tested for later)
        self.thumbnail = False

        self.curGroupLevel = 0 #track how deep we currently are in a group
        self.maxDepth = 64 #maximum number of group levels allowed
        self.curGroupAtLevelX = np.zeros((self.maxDepth,),dtype=np.int8) #track group at current level
        self.curGroupNameAtLevelX = '' #set the name of the root group

        self.curTagAtLevelX = np.zeros((self.maxDepth,),dtype=np.int8) #track tag number at the current level
        self.curTagName = '' #string of the current tag

        #lists that will contain scale information (pixel size)
        self.scale = []
        self.scaleUnit = []
        self.origin = []
        self.dataType = []

        #Temporary variables to keep in case a tag entry shows useful information in an array
        self.scale_temp = 0
        self.origin_temp = 0

        self.outputDic = {}
        self.allTags = {}
        
        self.parseHeader()
        
    def __del__(self):
        '''Destructor which also closes the file
        
        '''
        if(not self._on_memory and self.fid):
            if self.v:
                print('Closing input file: {}'.format(self.filename))
            self.fid.close()

    def __enter__(self):
        '''Implement python's with staement
        
        '''
        return self
        
    def __exit__(self,type,value,traceback):
        '''Implement python's with statment
        and close the file via __del__()
        '''
        self.__del__()
        return None
    
    def tell(self):
        '''Return the current position in the file. Switches mode based 
        on on_memory mode.
        
        '''
        if self._on_memory:
            return self._buffer_offset
        else:
            return self.fid.tell()

    def fromfile(self, *args, **kwargs):
        ''' Reads data from a file or memory map. Calls np.fromfile and 
        np.frombuffer depending on the on_memory mode of the fileDM.

        Parameters: it supports whatever frombuffer, fromfile support but it requires:
            dtype (numpy.dtype): Data type (object or string) to be read.
            count (int): Number of dtype items to be read.

        Returns:
            (list): A list of count dtype elements.
        
        '''
        
        if self._on_memory:
            if "dtype" not in kwargs:
                raise ValueError("In on_memory mode, reads require always a"
                                 " named dtype argument to be specified.")
            if "count" not in kwargs:
                raise ValueError("In on_memory mode, reads require always a"
                                 " named count argument to be specified.")

            dtype=np.dtype(kwargs["dtype"])
            count=int(kwargs["count"])
            kwargs["offset"]=self._buffer_offset
            self._buffer_offset+=int(dtype.itemsize)*count
            return np.frombuffer(*args, **kwargs)
        else:
            return np.fromfile(*args, **kwargs)

    def seek(self, fid, offset, from_what=0):
        '''Positions the reading head for fid. fid can be a file or memory map.
        Follows the same convention as file.seek

        Parameters:
            fid (file id): file or memory map.
            offset (int): number of bytes to move the head forward (positive value)
            or backwards (negative value).
            from_what (int): reference point to use in the head movement. 0:
            for beginning of the file (default behavior), 1: from the current
            head position, and 2: from the end of the file.
        
        Returns:
            None
        '''
        if self._on_memory:
            offset=int(offset)
            if from_what==0:
                self._buffer_offset=offset
            elif from_what==1:
                self._buffer_offset+=offset
            elif from_what==2:
                self._buffer_offset=self._buffer_size+offset-1
            else:
                raise ValueError("Unkown from_what value: {}".format(from_what))
            if self._buffer_offset<0:
                raise ValueError("Resulting head position cannot be negative.")
        else:
            return fid.seek(offset, from_what)


    def _validDM(self):
        '''Test whether a file is a valid DM3 or DM4 file and written
        in little endian format.
        
        '''
        output = True #output will stay == 1 if the file is a true DM4 file

        self.dmType = self.fromfile(self.fid,dtype=np.dtype('>u4'),count=1)[0] #file type: == 3 for DM3 or == 4 for DM4

        if self.v:
            print('validDM: DM file type numer = {}'.format(self.dmType))

        if self.dmType == 3:
            self.specialType = np.dtype('>u4') #uint32
        elif self.dmType == 4:
            self.specialType = np.dtype('>u8') #uint64
        else:
            raise IOError('File is not a valid DM3 or DM4')
            output = False

        self.fileSize = self.fromfile(self.fid,dtype=self.specialType,count=1)[0] #file size: real size - 24 bytes
        self.endianType = self.fromfile(self.fid,dtype=np.dtype('>u4'),count=1)[0] #endian type: 1 == little endian (Intel), 2 == big endian (old powerPC Mac)

        if self.endianType != 1:
            #print('File is not written Little Endian (PC) format and can not be read by this program.')
            raise IOError('File is not written Little Endian (PC) format and can not be read by this program.')
            output = False

        #Test file size for corruption. Note that DM3/DM4 file size is always off by 20/24 bytes from what is written in the header
        osSize = fileStats(self.filename).st_size
        if self.dmType == 3:
            if self.fileSize != osSize-20:
                pass
                #raise IOError('File size on disk ({}) does not match expected file size in header ({}). Invalid file.'.format(osSize, self.fileSize))
                #output = False
                #print('Warning: file size on disk ({}) does not match expected file size in header ({}).'.format(osSize, self.fileSize))
        elif self.dmType == 4:
            if self.fileSize != osSize-24:
                pass
                #raise IOError('File size on disk ({}) does not match expected file size in header ({}). Invalid file.'.format(osSize, self.fileSize))
                #output = False
                #print('Warning: file size on disk ({}) does not match expected file size in header ({}).'.format(osSize, self.fileSize))

        return output

    def parseHeader(self):
        '''Parse the header by reading the root tag group.
        This ensures the file pointer is in the correct place.
        
        '''
        #skip the bytes read by dmType
        if self.dmType == 3:
            self.seek(self.fid, 12,0)
        elif self.dmType == 4:
            self.seek(self.fid, 16,0)
        #Read the first root tag the same as any other group
        self._readTagGroup()
        
        #Check for thumbnail
        if(len(self.dataType) > 0): #check that any data set was found
            if( (self.dataType[0] == 23) & (self.dataShape[0] == 2) ):
                self.thumbnail = True
            else:
                self.thumbnail = False
        else: #this file only contains tags (such as a GTG file)
            self.thumbnail = False
        
        ''' Determine useful meta data UNTESTED
        self.metaData = {}
        for kk,ii in self.allTags.items():
            prefix1 = 'ImageList.{}.ImageTags.'.format(md.numObjects)
            prefix2 = 'ImageList.{}.ImageData.'.format(md.numObjects)
            pos1 = kk.find(prefix1)
            pos2 = kk.find(prefix2)
            if pos1 > -1:
                sub = kk[pos1+len(prefix):]
                self.metaData[sub] = ii
            elif pos2 > -1:
                sub = kk[pos2+len(prefix):]
                self.metaData[sub] = ii
            
            #Remove some unneeded keys
            for jj in list(self.metaData):
                if jj.find('frame sequence')>-1:
                    del self.metaData[jj]
                elif jj.find('Private')>-1:
                    del self.metaData[jj]
                elif jj.find('Reference Images')>-1:
                    del self.metaData[jj]
                elif jj.find('Frame.Intensity')>-1:
                    del self.metaData[jj]
                elif jj.find('Area.Transform')>-1:
                    del self.metaData[jj]
                elif jj.find('Parameters.Objects')>-1:
                    del self.metaData[jj]
                elif jj.find('Device.Parameters')>-1:
                    del self.metaData[jj]
        return metaData
        '''
        
    def _readTagGroup(self):
        '''Read a tag group in a DM file.
        
        '''
        self.curGroupLevel += 1
        #Check to see if the maximum group level is reached.
        if self.curGroupLevel > self.maxDepth:
            raise IOError('Maximum tag group depth of {} reached. This file is most likely corrupt.'.format(self.maxDepth))

        self.curGroupAtLevelX[self.curGroupLevel] = self.curGroupAtLevelX[self.curGroupLevel] + 1
        self.curTagAtLevelX[self.curGroupLevel] = 0
        self.fromfile(self.fid,dtype='<i1',count=2) #is open and is sorted?
        nTags = self.fromfile(self.fid,dtype=self.specialType,count=1)[0] #needs to be read as Big Endian (.byteswap() could also work)

        if self.v:
            print('Total number of root tags = {}'.format(nTags))

        #Iterate of the number of tag entries
        oldTotalTag = self.curGroupNameAtLevelX
        for ii in range(0,nTags):
            self._readTagEntry()

        #Go back down a level after reading all entries
        self.curGroupLevel -= 1
        self.curGroupNameAtLevelX = oldTotalTag

    def _readTagEntry(self):
        '''Read one entry in a tag group.
        
        '''
        dataType = self.fromfile(self.fid,dtype=np.dtype('>u1'),count=1)[0]

        #Record tag at this level
        self.curTagAtLevelX[self.curGroupLevel] += 1

        #get the tag
        lenTagLabel = self.fromfile(self.fid,dtype='>u2',count=1)[0]

        if self.v:
            print('_readTagEntry: dataType = {}, lenTagLabel = {}'.format(dataType,lenTagLabel))

        if lenTagLabel > 0:
            tagLabelBinary = self.fromfile(self.fid,dtype='<u1',count=lenTagLabel) #read as binary
            tagLabel = self._bin2str(tagLabelBinary)
            if self.v:
                print('_readTagEntry: tagLabel = {}'.format(tagLabel))
        else:
            tagLabel = str(self.curTagAtLevelX[self.curGroupLevel]) #unlabeled tag.

        #Save the current group name in case this is needed
        oldGroupName = self.curGroupNameAtLevelX

        if dataType == 21:
            #This tag entry contains data
            self.curTagName = tagLabel #save its name
            self._readTagType()
        else:
            #This is a nested tag group
            self.curGroupNameAtLevelX += '.' + tagLabel #add to group names

            #An unknown part of the DM4 tags
            if self.dmType == 4:
                temp1 = self.fromfile(self.fid,dtype=self.specialType,count=1)[0]

            self._readTagGroup()

        self.curGroupNameAtLevelX = oldGroupName

    def _readTagType(self):
        '''Determine the type of tag: Regular data, string, struct, or array.
        
        '''
        #Need to read 8 bytes before %%%% delimiater. Unknown part of DM4 tag structure
        if self.dmType == 4:
            temp1 = self.fromfile(self.fid,dtype=self.specialType,count=1)[0]

        delim = self.fromfile(self.fid,dtype='<i1',count=4)
        assert((delim == 37).all()) #delim has to be [37,37,37,37] which is %%%% in ASCII.
        if self.v:
            print('_readTagType: should be %%%% = {}'.format(self._bin2str(delim)))

        nInTag = self.fromfile(self.fid,dtype=self.specialType,count=1)[0] #nInTag: unnecessary redundant info

        #Determine the type of the data in the tag
        #specifies data type: int8, uint16, float32, etc.
        encodedType = self.fromfile(self.fid,dtype=self.specialType,count=1)[0] #big endian

        etSize = self._encodedTypeSize(encodedType)

        if etSize > 0:
            #regular data. Read it and store it with the tag name
            if self.v:
                print('regular')
            self._storeTag(self.curTagName, self._readNativeData(encodedType))
        elif encodedType == 18: #string
            if self.v:
                print('string')
            stringSize = self.fromfile(self.fid,dtype='>u4',count=1)[0]
            #strtemp = '' #in case stringSize == 0
            strTempBin = self.fromfile(self.fid,dtype='<u1',count=stringSize) #read as uint8 little endian
            strTemp = self._bin2str(strTempBin)
            self._storeTag(self.curTagName,strTemp)
        elif encodedType == 15: #struct
            #This does not work for field names that are non-zero. This is uncommon
            if self.v:
                print('struct')
            structTypes = self._readStructTypes()
            structs = self._readStructData(structTypes)
            self._storeTag(self.curTagName,structs)
        elif encodedType == 20: #array
            #The array data is not read. It will be read later if needed
            if self.v:
                print('array')
            arrayTypes = self._readArrayTypes() #could be recursive if array contains array(s)
            arrInfo = self._readArrayData(arrayTypes) #only info of the array is read. It is read later if needed
            self._storeTag(self.curTagName,arrInfo)

    def _bin2str(self,bin):
        '''Utility function to convert a numpy array of binary values to a python string.
        
        '''
        return ''.join([chr(item) for item in bin])

    def _encodedTypeSize(self, encodedType):
        '''Return the number of bytes in a data type for the encodings used by DM.
        Constants for the different encoded data types used in DM files as as follows:
            SHORT   = 2
            LONG    = 3
            USHORT  = 4
            ULONG   = 5
            FLOAT   = 6
            DOUBLE  = 7
            BOOLEAN = 8
            CHAR    = 9
            OCTET   = 10
            uint64  = 12
            -1 will signal an unlisted type
            
        Parameters:
            encodedType (int): The type value read from the header
            
        Returns:
            (int): Number of bytes this type uses.
        '''
        if encodedType == 0:
            return 0
        elif (encodedType == 8) | (encodedType == 9) | (encodedType == 10):
            return 1 #1 byte
        elif (encodedType == 2) | (encodedType == 4):
            return 2 #2 bytes
        elif (encodedType == 3) | (encodedType == 5) | (encodedType == 6):
            return 4 #4 bytes
        elif (encodedType == 7) | (encodedType == 12):
            return 8 #8 bytes
        else:
            return -1

    def _encodedTypeDtype(self,encodedType):
        '''Translate the encodings used by DM to numpy dtypes according to:
            SHORT   = 2
            LONG    = 3
            USHORT  = 4
            ULONG   = 5
            FLOAT   = 6
            DOUBLE  = 7
            BOOLEAN = 8
            CHAR    = 9
            OCTET   = 10
            uint64  = 12
            -1 will signal an unlisted type
            
        Parameters:
            encodedType (int): The type value read from the header
            
        Returns:
            (numpy dtype): The Numpy dtype corresponding to the DM encoded value.
        
        '''
        if encodedType == 2:
            return np.dtype('<i2')
        elif encodedType == 3:
            return np.dtype('<i4')
        elif encodedType == 4:
            return np.dtype('<u2')
        elif encodedType == 5:
            return np.dtype('<u4')
        elif encodedType == 6:
            return np.dtype('<f4')
        elif encodedType == 7:
            return np.dtype('<f8')
        elif encodedType == 8:
            return np.dtype('<u1')
        elif encodedType == 9:
            return np.dtype('<u1')
        elif encodedType == 10:
            return np.dtype('<u1')
        elif encodedType == 12:
            return np.dtype('<u8')
        else:
            return -1

    def _readStructTypes(self):
        '''Analyze the types of data in a struct.
        
        '''
        structNameLength = self.fromfile(self.fid,count=1,dtype=self.specialType)[0] #this is not needed
        nFields = self.fromfile(self.fid,count=1,dtype=self.specialType)[0]
        if self.v:
            print('_readStructTypes: nFields = {}'.format(nFields))

        if(nFields > 100):
            raise RuntimeError('Too many fields in a struct.')

        fieldTypes = np.zeros(nFields)
        for ii in range(0,nFields):
            aa = self.fromfile(self.fid,dtype=self.specialType,count=2) #nameLength, fieldType
            nameLength = aa[0] #not used currently
            fieldTypes[ii] = aa[1]
        return fieldTypes

    def _readStructData(self,structTypes):
        '''Read the data in a struct.
        
        Parameters:
            structTypes (ndarray): 1D array containing fieldTypes
        
        Returns:
            (ndarray): 1D array of data
        '''
        struct = np.zeros(structTypes.shape[0])
        for ii, encodedType in enumerate(structTypes):
            etSize = self._encodedTypeSize(encodedType) #the size of the data type
            struct[ii] = self._readNativeData(encodedType) #read this type of data
        return struct

    def _readNativeData(self,encodedType):
        '''Reads ordinary data types in tags according to:
            SHORT (int16)   = 2
            LONG (int32)    = 3
            USHORT (uint16)  = 4
            ULONG (uint32)   = 5
            FLOAT (float32)  = 6
            DOUBLE (float64)  = 7
            BOOLEAN (bool) = 8
            CHAR (uint8 character) = 9
            OCTET (??)  = 10
            UINT64 (uint64) = 11
            
        Parameters:
            encodedType (int): Encoded type value from DM header
        
        Returns:
            (int or ndarray): The value(s) read in.
        '''
        if encodedType == 2:
            val = self.fromfile(self.fid,count=1,dtype='<i2')[0]
        elif encodedType == 3:
            val = self.fromfile(self.fid,count=1,dtype='<i4')[0]
        elif encodedType == 4:
            val = self.fromfile(self.fid,count=1,dtype='<u2')[0]
        elif encodedType == 5:
            val = self.fromfile(self.fid,count=1,dtype='<u4')[0]
        elif encodedType == 6:
            val = self.fromfile(self.fid,count=1,dtype='<f4')[0]
        elif encodedType == 7:
            val = self.fromfile(self.fid,count=1,dtype='<f8')[0]
        elif encodedType == 8: #matlab uchar
            val = self.fromfile(self.fid,count=1,dtype='<u1')[0] #return character or number?
            if self.v:
                print('_readNativeData untested type, val: {}, {}'.format(encodedType,val))
        elif encodedType == 9: #matlab *char
            val = self.fromfile(self.fid,count=1,dtype='<i1')[0] #return character or number?
            if self.v:
                print('_readNativeData untested type, val: {}, {}'.format(encodedType,val))
        elif encodedType == 10: #matlab *char
            val = self.fromfile(self.fid,count=1,dtype='<i1')[0]
            if self.v:
                print('_readNativeData untested type, val: {}, {}'.format(encodedType,val))
        elif encodedType == 11:
            val = self.fromfile(self.fid,count=1,dtype='<u8')[0]
        elif encodedType == 12:
            val = self.fromfile(self.fid,count=1,dtype='<u8')[0] #unknown type, but this works
        else:
            print('_readNativeData unknown data type: {}'.format(encodedType))
            raise

        if self.v:
            print('_readNativeData: encodedType == {} and val = {}'.format(encodedType, val))

        return val
    def _readArrayTypes(self):
        '''Analyze the types of data in an array.
        
        '''
        arrayType = self.fromfile(self.fid,dtype=self.specialType,count=1)[0]

        itemTypes = []

        if arrayType == 15:
            #nested Struct
            itemTypes = self._readStructTypes()
        elif arrayType == 20:
            #Nested array
            itemTypes = _readArrayTypes()
        else:
            itemTypes.append(arrayType)
        if self.v:
            print('_readArrayTypes: itemTypes = {}'.format(itemTypes))
        return itemTypes

    def _readArrayData(self,arrayTypes):
        '''Read information in an array based on the types provided.
        Binary data (i.e. image/spectra data) is skipped in order to 
        save memory. These are read later using getDataset() or
        getSlice as needed.
        
        Parameters:
            arrayTypes (ndarray or tuple): The type of array data to read
            
        Returns:
            (str): A string containing the key value pair of this tag
        
        '''

        #The number of elements in the array
        arraySize = self.fromfile(self.fid,count=1,dtype=self.specialType)[0]

        if self.v:
            print('_readArrayData: arraySize, arrayTypes = {}, {}'.format(arraySize,arrayTypes))

        #Everything used to calculate the bufSize is not needed anymore. THis can be removed after testing
        itemSize = 0
        for encodedType in arrayTypes:
            if self.v:
                print('_readArrayData: encodedType = {}'.format(encodedType))
            etSize = self._encodedTypeSize(encodedType)
            itemSize += etSize

        bufSize = arraySize * itemSize
        bufSize = bufSize.astype('<u8') #change to an integer

        if self.v:
            print('_readArrayData: arraySize, itemSize = {}, {}'.format(arraySize, itemSize))

        if self.curTagName == 'Data':
            #This is a binary array. Save its location to read later if needed
            self._storeTag(self.curTagName + '.arraySize', bufSize)
            self._storeTag(self.curTagName + '.arrayOffset', self.tell())
            self._storeTag(self.curTagName + '.arrayType', encodedType)
            self.seek(self.fid, bufSize.astype('<u8'),1) #advance the pointer by bufsize from current position
            arrOut = 'Data unread. Encoded type = {}'.format(encodedType)
        elif bufSize < 1e3: #set an upper limit on the size of array that will be read in as a string
            #treat as a string
            for encodedType in arrayTypes:
                stringData = self.fromfile(self.fid,count=arraySize,dtype=self._encodedTypeDtype(encodedType))
                arrOut = self._bin2str(stringData)

            #This is the old way to read this in. Its not really correct though.
            #stringData = self.bin2str(self.fromfile(self.fid,count=bufSize,dtype='<u1'))
            #arrOut = stringData.replace('\x00','') #remove all spaces from the string data

            #Catch useful tags for images and spectra (nm, eV, etc.)
            fullTagName = self.curGroupNameAtLevelX + '.' + self.curTagName
            if((fullTagName.find('Dimension') > -1) & (fullTagName.find('Units') > -1) ):# & (self.numObjects > 0)):
                self.scale.append(self.scale_temp)
                self.scaleUnit.append(arrOut)
                self.origin.append(self.origin_temp)
        else:
            self._storeTag(self.curTagName + '.arraySize', bufSize)
            self._storeTag(self.curTagName + '.arrayOffset', self.tell())
            self._storeTag(self.curTagName + '.arrayType', encodedType)
            self.seek(self.fid, bufSize.astype('<u8'),1) #advance the pointer by bufsize from current position
            arrOut = 'Array unread. Encoded type = {}'.format(encodedType)

        return arrOut

    def _storeTag(self,curTagName,curTagValue):
        '''Builds the full tag name and key/value pair as text. Also calls another
        function to catch useful tags and values. Also saves all tags in a dictionary.
        
        Parameters:
            curTagName (str): The Tag name; a key
            curTagValue (): This can be many different type of objets
            like ints, floats, stings.
        
        '''
        #Build the full tag name (key) and add the tag value
        if self.v:
            print('_storeTag: curTagName, curTagValue = {}, {}'.format(curTagName,curTagValue))
        totalTag = self.curGroupNameAtLevelX + '.' + '{}'.format(curTagName) #+ '= {}'.format(curTagValue)

        self._catchUsefulTags(totalTag,curTagName,curTagValue)

        self.allTags[totalTag] = curTagValue #this needs to be done better.

        return(totalTag)

    def _catchUsefulTags(self,totalTag,curTagName,curTagValue):
        '''Find interesting keys and keep their values for later. This is separate from _storeTag
        so that it is easy to find and modify.
        
        Parameters:
            totalTag (str): The complete tag as a string
            curTagName (str): The tag name; the tag key
            curTagValue (): can be many different types of objects
        
        '''

        #Save that a useful object has been found
        if totalTag.find('ImageData.Calibrations.Dimension.1.Scale')>-1:
            self.numObjects += 1 #data is contained in this file

        if curTagName.find('Data.arraySize')>-1:
            self.dataSize.append(curTagValue)
        elif curTagName.find('Data.arrayOffset') >-1:
            self.dataOffset.append(curTagValue)
        elif curTagName.find('DataType')>-1:
            self.dataType.append(curTagValue)
        elif totalTag.find('Dimensions.1')>-1:
            self.xSize.append(curTagValue)
            self.ySize.append(1)
            self.zSize.append(1)
            self.zSize2.append(1)
            self.dataShape.append(1) # indicate as at least 1D data
        elif totalTag.find('Dimensions.2')>-1:
            self.ySize[-1] = curTagValue #OR self.ysize[self.numObjects] = self.curTagValue
            self.dataShape[-1] = 2 # indicate as at least 2D data
        elif totalTag.find('Dimensions.3')>-1:
            self.zSize[-1] = curTagValue
            self.dataShape[-1] = 3 # indicate as at least 3D data
        elif totalTag.find('Dimensions.4')>-1:
            self.zSize2[-1] = curTagValue
            self.dataShape[-1] = 4 # indicate as at least 3D data
        elif (totalTag.find('Dimension.')>-1) & (totalTag.find('.Scale')>-1):
            self.scale_temp = curTagValue
        elif (totalTag.find('Dimension.')>-1) & (totalTag.find('.Origin')>-1):
            self.origin_temp = curTagValue
        else:
            pass

    def writeTags(self):
        '''Write  out all tags as human readable text to a text file
        in the same directory and with a the same name as the DM file.
        
        '''
        fnameOutPrefix = self.filename.split('.dm3')[0]
        try:
            #open a text file to write out the tags
            with open(fnameOutPrefix+'_tags.txt','w') as fidOut:
                for nn in sorted(self.allTags):
                    try:
                        oo = '{} = {}'.format(nn,str(self.allTags[nn]))
                        fidOut.write(oo)
                    except:
                        fidOut.write('{} = dm.py error'.format(nn))
                    fidOut.write('\n')
        except NameError:
            print("Issue opening tags output file.")
            raise
        except:
            raise

    def _checkIndex(self, i):
        '''Check index i for sanity, otherwise raise Exception.

        Parameters:
            i (int):    Index.
        
        Raises:
            IndexError
        '''

        # check type
        if not isinstance(i, int):
            raise TypeError('index supposed to be integer')

        # check whether in range
        if i < 0 or i > self.numObjects:
            raise IndexError('Index out of range, trying to access element {} of {} valid elements'.format(i+1, self.head['ValidNumberElements']))

        return

    def _DM2NPDataType(self, dd):
        '''Convert the DM data type value into a numpy dtype.
        
        Parameters:
            dd (int): The value encoded in the DM file header.
            
        Returns:
            (numpy dtype)
        '''
        if dd == 6:
            return np.uint8
        elif dd == 10:
            return np.uint16
        elif dd == 11:
            return np.uint32
        elif dd == 9:
            return np.int8
        elif dd == 1:
            return np.int16
        elif dd == 7:
            return np.int32
        elif dd == 2:
            return np.float32
        elif dd == 12:
            return np.float64
        #elif dd == 14: #this is supposed to be bit1 in matlab, but Im not sure what that translates to in numpy
        #    return np.uint8 #bit1 ??
        elif dd == 3:
            return np.complex64
        elif dd == 13:
            return np.complex128
        elif dd == 23:
            raise IOError('RGB data type is not supported yet.')
            #return np.uint8
        else:
            print("4321")
            raise IOError('Unsupported binary data type during conversion to numpy dtype. DM dataType == {}'.format(dd))

    def getDataset(self, index):
        '''Retrieve a dataset from the DM file.
        
        Note:
            Most DM3 and DM4 files contain a small "thumbnail"
            as the first dataset written as RGB data. This
            function ignores that dataset if it exists. To
            retrieve the thumbnail use the getThumbnail()
            function
            
        Parameters:
            index (int): The number of the data set to retrieve ignoring the thumbnail.
            If a thumbnail exists then inedx = 0 corresponds to second data
            set in a DM file. 
            
        Returns:
            (dict): A dictionary of the data and meta data. The data is associated
            with the 'data' key in the dictionary.
            
        '''
        #The first dataset is usually a thumbnail. Test for this and skip the thumbnail automatically
        if self.numObjects == 1:
            ii = index
        else:
            ii = index + 1

        #Check that the dataset exists.
        try:
            self._checkIndex(ii)
        except:
            raise

        self.seek(self.fid, self.dataOffset[ii],0) #Seek to start of dataset from beginning of the file

        outputDict = {}

        outputDict['filename'] = osBasename(self.filename)

        #Parse the dataset to see what type it is (image, image series, spectra, etc.)
        if self.xSize[ii] > 0:
            pixelCount = self.xSize[ii]*self.ySize[ii]*self.zSize[ii]*self.zSize2[ii]
            jj = 0 #counter to determine where the first scale value starts
            for nn in self.dataShape[0:ii]:
                    jj += nn #sum up all number of dimensions for previous datasets
            #if self.dataType == 23: #RGB image(s)
            #    temp = self.fromfile(self.fid,count=pixelCount,dtype=np.uint8).reshape(self.ysize[ii],self.xsize[ii])
            if self.zSize[ii] == 1: #2D data
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.ySize[ii],self.xSize[ii]))
                outputDict['pixelUnit'] = self.scaleUnit[jj:jj+self.dataShape[ii]][::-1] #need to reverse the order to match the C-ordering of the data
                outputDict['pixelSize'] = self.scale[jj:jj+self.dataShape[ii]][::-1]
                outputDict['pixelOrigin'] = self.origin[jj:jj+self.dataShape[ii]][::-1]
            elif self.zSize2[ii] > 1: #4D data
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.zSize2[ii],self.zSize[ii],self.ySize[ii],self.xSize[ii]))
                outputDict['pixelUnit'] = self.scaleUnit[jj:jj+self.dataShape[ii]][::-1] #need to reverse the order to match the C-ordering of the data
                outputDict['pixelSize'] = self.scale[jj:jj+self.dataShape[ii]][::-1]
                outputDict['pixelOrigin'] = self.origin[jj:jj+self.dataShape[ii]][::-1]
            else: #3D array
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.zSize[ii],self.ySize[ii],self.xSize[ii]))
                outputDict['pixelUnit'] = self.scaleUnit[jj:jj+self.dataShape[ii]][::-1] #need to reverse the order to match the C-ordering of the data
                outputDict['pixelSize'] = self.scale[jj:jj+self.dataShape[ii]][::-1]
                outputDict['pixelOrigin'] = self.origin[jj:jj+self.dataShape[ii]][::-1]

        return outputDict
    
    def getSlice(self,index,sliceZ,sliceZ2=0):
        '''Retrieve a slice of a dataset from the DM file. The data set will have a shape according to
        3D = [sliceZ,Y,X] or 4D: [sliceZ2,sliceZ,Y,X]
        
        Note: Most DM3 and DM4 files contain a small "thumbnail" as the first dataset written as RGB data. This function ignores that dataset if it exists. To retrieve the thumbnail use the getThumbnail() function.
        
        Warning: DM4 files with 4D data sets are written as [X,Y,Z1,Z2]. This code currently gets the [X,Y] slice. 
        Getting the [Z1,Z2] slice is not yet implemented.
        
        Parameters:
            index (int): The number of the dataset in the DM file.
            sliceZ (int): The slice to get along the first dimension (C-ordering) for 3D datasets or the 2nd dimensions for 
        
        Keywords:
            sliceZ2 (int): For 4D dataset
    
        Returns:
            (dict): a dictionary containing meta data and the data.
        '''
        #The first dataset is usually a thumbnail. Test for this and skip the thumbnail automatically
        if self.numObjects == 1:
            ii = index
        else:
            ii = index + 1

        #Check that the dataset exists.
        try:
            self._checkIndex(ii)
        except:
            raise
        
        # Check sliceZ and sliceZ2 are within the data arrray size bounds
        if sliceZ > (self.zSize[ii]-1):
            raise IndexError('Index out of range, trying to access element {} of {} valid elements'.format(sliceZ, self.zSize))
        if sliceZ2 > (self.zSize2[ii]-1):
            raise IndexError('Index out of range, trying to access element {} of {} valid elements'.format(sliceZ2, self.zSize2))
        
        self.seek(self.fid, self.dataOffset[ii],0) #Seek to start of dataset from beginning of the file
        
        outputDict = {}
        outputDict['filename'] = osBasename(self.filename)
        
        #Parse the dataset to see what type it is (image, 3D image series, spectra, 4D, etc.)
        if self.xSize[ii] > 0:
            #determine the number of bytes to skip
            pixelCount = self.xSize[ii]*self.ySize[ii]
            byteCount = pixelCount * np.dtype(self._DM2NPDataType(self.dataType[ii])).itemsize
            jj = 0 #counter to determine where the first scale value starts
            for nn in self.dataShape[0:ii]:
                    jj += nn #sum up all number of dimensions for previous datasets
            if self.zSize[ii] == 1: #2D data
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.ySize[ii],self.xSize[ii]))
            elif self.zSize2[ii] > 1: #4D data
                self.fid.seek(sliceZ*sliceZ2*byteCount,1) #skip ahead from current position
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.ySize[ii],self.xSize[ii]))
            else: #3D array
                self.fid.seek(sliceZ*byteCount,1) #skip ahead from current position
                outputDict['data'] = self.fromfile(self.fid,count=pixelCount,dtype=self._DM2NPDataType(self.dataType[ii])).reshape((self.ySize[ii],self.xSize[ii]))
            
            #Return the proper meta data for this one image
            outputDict['pixelUnit'] = self.scaleUnit[jj:jj+2][::-1] #need to reverse the order to match the C-ordering of the data
            outputDict['pixelSize'] = self.scale[jj:jj+2][::-1]
            outputDict['pixelOrigin'] = self.origin[jj:jj+2][::-1]
            
        return outputDict
    
    def _readRGB(self,xSizeRGB,ySizeRGB):
        '''Read in a uint8 type array with [Red,green,blue,alpha] channels.
        
        '''
        return self.fromfile(self.fid,count=xSizeRGB*ySizeRGB*4,dtype='<u1').reshape(xSizeRGB,ySizeRGB,4)

    def getThumbnail(self):
        '''Read the thumbnail saved as the first dataset in the DM file as an RGB array.
        This is not fully tested. Be careful using this.
        
        Returns:
            {ndarrya}: numpy array of size [3,Y,X] which is an RGB thumbnail.
        '''
        self.seek(self.fid, self.dataOffset[0],0)
        return self._readRGB(self.ySize[0],self.xSize[0])

def dmReader(filename,dSetNum=0,verbose=False):
    '''A simple function to parse the file and read the requested dataset.
    Most users will want to use this function to simplify reading data
    directly into memory.
    
    Parameters:
        filename (str) : The filename to open and read into memory
        dSetNum (int) : The number of the data set to read. Almost always should be = 0. Default = 0
        verbose (bool) : Allow extra printing to see file internals. Default = False

    Returns:
        (dict) : A dictionary of keys where the data is in the 'data' key. Other metadata is contained in other named keys such as 'pixelSize'
    
     Example:
        Load all data from a single image dm3 file and display it:
            >>> from ncempy.io import dm
            >>> im0 = dm.dmReader('filename.dm3')
            >>> plt.imshow(im0['data']) #show the single image from the data file
    '''
    with fileDM(filename,verbose) as f1: #open the file and init the class
        im1 = f1.getDataset(dSetNum) #get the requested dataset (first by default)
    
    return im1 #return the dataset and metadata as a dictionary



'''
Tests for the emd io module.
'''

import unittest
import os
import os.path
import numpy as np

import ncempy.io.emd

class test_emd(unittest.TestCase):
    '''
    Test the EMD io module
    '''
    
    def test_init_emd(self):
        
        # wrong argument type
        with self.assertRaises(TypeError):
            femd = ncempy.io.emd.fileEMD(42)
        
        # non existing file in readonly
        with self.assertRaises(IOError):
            femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/doesnotexist.emd', readonly=True)
            
        # impossible file for read/write
        with self.assertRaises(IOError):
            femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/output/output.emd')
            
        # open existing file
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/Au_SAED_D910mm_20x_at_800.emd')
        # open existing file readonly
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/Au_SAED_D910mm_20x_at_800.emd', readonly=True)
        # open nonexisting file for writing
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/test.emd')


    def test_data_manipulation(self):
    
        # open testfiles
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/Au_SAED_D910mm_20x_at_800.emd', readonly=True)
        if os.path.isfile('ncempy/test/resources/output/copy.emd'):
            os.remove('ncempy/test/resources/output/copy.emd')
        femd2 = ncempy.io.emd.fileEMD('ncempy/test/resources/output/copy.emd')
        
        # test error handling in get_emdgroup
        with self.assertRaises(TypeError):
            data, dims = femd.get_emdgroup(femd.file_hdl)
        
        # working
        data, dims = femd.get_emdgroup(femd.list_emds[0])
        
        # test error handling in put_emdgroup
        with self.assertRaises(TypeError):
            femd2.put_emdgroup(42, data,  dims)
            
        with self.assertRaises(TypeError):
            femd2.put_emdgroup('Au_SAED_D910mm_20x_at_800', 42, dims)
        
        with self.assertRaises(TypeError):
            femd2.put_emdgroup('Au_SAED_D910mm_20x_at_800', data, dims[0:1])
        
        # working
        self.assertIsNotNone(femd2.put_emdgroup('Au_SAED_D910mm_20x_at_800', data, dims))
        
        # try to write a readonly file
        femd2 = ncempy.io.emd.fileEMD('ncempy/test/resources/output/copy.emd', readonly=True)
        self.assertIsNone(femd2.put_emdgroup('Au_SAED_D910mm_20x_at_800', data, dims))
        
        del femd, femd2
        
        # write a testfile
        if os.path.isfile('ncempy/test/resources/output/test2.emd'):
            os.remove('ncempy/test/resources/output/test2.emd')
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/test2.emd')
        data = np.random.rand(512,512,100)
        dims = ( (np.array(range(512)), 'x', '[px]'),
                 (np.array(range(512)), 'y', '[px]'),
                 (np.linspace(0.0, 3.14, num=100), 'angle','[rad]') )
        self.assertIsNotNone(femd.put_emdgroup('dataset_1', data, dims))
        
        dim = (np.array(range(100)), 'number', '[]')
        femd.write_dim('dim3_number', dim, femd.list_emds[0])
        
        femd.put_comment('file created, filled with random numbers')
        
        # try to overwrite
        self.assertIsNone(femd.put_emdgroup('dataset_1', data, dims))
        
        
    def test_comments(self):
    
        # create a file for comments
        if os.path.isfile('ncempy/test/resources/output/comments.emd'):
            os.remove('ncempy/test/resources/output/comments.emd')
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/comments.emd')
        
        with self.assertRaises(TypeError):
            femd.put_comment(42)
        
        femd.put_comment('file created')
        femd.put_comment('how long does it take until the second comment is issued?')
        
        femd.put_comment('something happened', 'today')
        femd.put_comment('even more happened', 'today')

# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
Module to calculate radial profiles.

The **settings** dict used for the evaluation has the following entries:

    * lmax_r (int):    Radius used for local maximum detection in [px].
    * lmax_thresh (int):    Threshold for local maximum detection in [intensity].
    * lmax_cinit (int, int):    Initial guess for center in [px].
    * lmax_range (float, float):    r range to cinit used to filter local maxima [dims].
    * ns (int, ...):    Distortion orders to correct.
    * fit_rrange (float, float):    r range used to fit radial profile [dims].
    * back_xs (float, ...):    Fixpoints for fitting background [dims].
    * back_xswidth (float):    Range around fixpoints to take into account [dims].
    * back_init (float):    Initial guess for subtracting background.
    * fit_funcs (str, ...):    List of functions to model radial profile.
    * fit_init (float, ...):    Initial guess for fitting.

    optional (set to None to use defaults):
    
        * plt_imgminmax (float, float):    Relative range to plot the img.
        * rad_rmax (float):    Max r to be used in radial profile [dims].
        * rad_dr (float): Stepsize for r-axis in radial profile [dims].
        * rad_sigma (float): Sigma for Gaussian used as kernel density estimator [dims].
        * mask (np.ndarray): Binary image as img, 0 for pixels to exclude.
        * fit_maxfev (int): Maxfev forwarded to scipy optimize.

'''

import copy
import numpy as np
import scipy.ndimage.filters
import scipy.interpolate
import matplotlib.pyplot as plt

import ncempy.algo.math
import ncempy.algo.distortion

def calc_polarcoords ( center, dims, ns=None, dists=None ):
    '''Calculate the polar coordinates for an image of given shape.
    
    Center is assumed to be in real coordinates (if not just fake the dims).
    Distortions are corrected if ns and corresponding dists are given.
    
    Parameters:
        center (np.ndarray/tuple):    Center of polar coordinate system.
        dims (tuple):    Tuple of dimensions.
        ns (tuple):    List of distortion orders to correct for.
        dists (np.ndarray):    Parameters for distortions.
    
    Returns:
        (tuple):    Polarcoordinates (r, theta) of polar coordinate system as two np.ndarrays with same dimensions as original image.
        
    '''
    
    # check input
    try:
        # check center 
        center = np.array(center)
        center = np.reshape(center, 2)
        
        # check if enough dims availabel
        assert(len(dims)>=2)
        assert(len(dims[0])==3)
        
        # check orders
        if not ns is None:
            assert(len(ns)>=1)
        
            # check dists
            assert(dists.shape[0] == len(ns)*2+1)
            
    except:
        raise TypeError('Something wrong with the input!')
    
    # create coordinates
    xx, yy = np.meshgrid( dims[0][0], dims[1][0], indexing='ij' )
    
    # calculate polar coordinate system
    rs = np.sqrt( np.square(xx-center[0]) + np.square(yy-center[1]) )
    thes = np.arctan2(yy-center[1], xx-center[0])
    
    # correct for distortions
    if not ns is None:
        for i in range(len(ns)):
            rs /= ncempy.algo.distortion.rad_dis(thes, dists[i*2+1], dists[i*2+2], ns[i])    
    
    return rs, thes
    
      
def correct_distortion( img, dims, center, ns, dists ):
    '''Give corrected version of img with respect to distortions.
    
    Parameters:
        img (np.ndarray):    Image (2D).
        dims (tuple):    Dimensions tuple.
        center (np.ndarray/tuple):    Center to be used.
        ns (tuple):    List of distortion orders.
        dists (np.ndarray):    Distortion parameters.
        
    Returns:
        (np.ndarray):   Corrected image.
        
    '''
    
    # check input
    try:
        assert(isinstance(img, np.ndarray))
    
        # check center 
        center = np.array(center)
        center = np.reshape(center, 2)
        
        # check if enough dims availabel
        assert(len(dims)>=2)
        assert(len(dims[0])==3)
        
        # check orders
        assert(len(ns)>=1)
        
        # check dists
        assert(dists.shape[0] == len(ns)*2+1)
            
    except:
        raise TypeError('Something wrong with the input!')
    
    rs, thes = calc_polarcoords (center, dims )
    
    # anti distort
    for i in range(len(ns)):
        rs *= ncempy.algo.distortion.rad_dis(thes, dists[i*2+1], dists[i*2+2], ns[i]) 
    
    dis_xx = rs*np.cos(thes)+center[0]
    dis_yy = rs*np.sin(thes)+center[1]
    
    f = scipy.interpolate.RectBivariateSpline(dims[0][0], dims[1][0], img)

    return f.ev(dis_xx, dis_yy)
        
    
def calc_radialprofile( img, rs, rMax, dr, rsigma, mask=None):
    '''Calculate the radial profile using Gaussian kernel density estimator.
    
    It is suggested to use an rMax such that all directions are still in the image, otherwise outer areas will contribute differently and lead to different signal-to-noise. A value of dr corresponding to 1/10 px and rsigma corresponding to 1 px are good parameters.
    
    Parameters:
        img (np.ndarray):    Image to take radial profile of intensity from.
        rs (np.ndarray):    Array containing the radial distance, same shape as img.
        rMax (float):      Maximum radial distance used for the radial profile.
        dr (float):    Stepsize for r-axis of radial distance.
        rsigma (float):    Sigma for Gaussian used as kernel density estimator.
        mask (np.ndarray):    Mask image, values at 0 are excluded from radial profile.
    
    Returns:
        (tuple):    Tuple of radial and intensity axes.
        
    '''
    
    # check input
    try:
        assert(isinstance(img, np.ndarray))
        assert(isinstance(rs, np.ndarray))
        assert(np.array_equal(img.shape, rs.shape))
        
        rMax = float(rMax)
        dr = float(dr)
        rsigma = float(rsigma)
        
        if not mask is None:
            assert(isinstance(mask, np.ndarray))
            assert(np.array_equal(img.shape, mask.shape))
        
    except:
        raise TypeError('Something wrong with input.')
    
    # process mask
    if mask is None:
        mask = np.ones(img.shape)
    mask = mask.astype('float64')
    mask[np.where( mask > 0 )] = 1
    mask[np.where( mask == 0 )] = np.nan
    
    
    # prepare radial axis for hist  
    rBins = np.arange(0, rMax, dr)
    
    radialIndex = np.round(rs/dr)+1
    rBinMax = len(rBins)
    sel = (radialIndex<=rBinMax)
    
    sel = np.logical_and(sel, np.logical_not(np.isnan(mask)) )
    
    signal = np.histogram(rs[sel], rBins, weights=img[sel])
    count = np.histogram(rs[sel], rBins, weights=np.ones(img[sel].shape))

    signal_sm = scipy.ndimage.filters.gaussian_filter1d(signal[0], rsigma/dr)
    count_sm = scipy.ndimage.filters.gaussian_filter1d(count[0], rsigma/dr)
    
    # masked regions lead to 0 in count_sm, divide produces nans, just ignore the warning    
    old_err_state = np.seterr(divide='ignore',invalid='ignore')    
    signal_sm = np.divide(signal_sm,count_sm)
    np.seterr(**old_err_state)    
        
    return rBins[:-1], signal_sm
    
    
def plot_radialprofile( r, intens, dims, show=False ):
    '''Plot radial profile.
    
    Parameters:
        r (np.ndarray):    r-axis of radial profile.
        intens (np.ndarray):    Intensity-axis of radial profile.
        dims (tuple):    Dimensions of original image to read out units.
        show (bool):    Set to directly show plot interactively.
    
    Returns:
        (np.ndarray):    Image of the plot.
        
    '''
    
    try:
        # check data
        assert(isinstance(r, np.ndarray))
        assert(isinstance(intens, np.ndarray))
        assert(np.array_equal(r.shape, intens.shape))
        
        # check if dims availabel
        assert(len(dims)>=1)
        assert(len(dims[0])==3)

    except:
        raise TypeError('Something wrong with the input!')
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    ax.plot(r, intens, 'r-')
    
    # labels
    ax.set_xlabel('r /{}'.format(dims[0][2]))
    ax.set_ylabel('I /[a.u.]')
        
    if show:
        plt.show(block=False)
    
    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    return plot


def residuals_fit( param, r, intens, funcs ):
    '''Residual function to fit radial profile with flexibility.
    
    The arguments for the single functions are automatically cut from the fit parameters.
    
    Parameters:
        param (np.ndarray):    Fit parameters.
        r (np.ndarray):    r-axis.
        intens (np.ndarray):    Intensity-axis.
        funcs (tuple):    List of functions to include.
        
    Returns:
        (np.ndarray):    Residuals.
        
    '''

    return intens - ncempy.algo.math.sum_functions(r, funcs, param)
    
    
def fit_radialprofile( r, intens, funcs, init_guess, maxfev=None ):
    '''Fit the radial profile.
    
    Convenience wrapper for fitting.
    
    Parameters:
        r (np.ndarray):    r-axis of radial profile.
        intens (np.ndarray):    Intensity-axis of radial profile.
        funcs (tuple):    List of functions.
        init_guess (np.ndarray):    Initial guess for parameters of functions in funcs.
        
    Returns:
        (np.ndarray):    Optimized parameters.
        
    '''    
    
    try:
        # check data
        assert(isinstance(r, np.ndarray))
        assert(isinstance(intens, np.ndarray))
        assert(np.array_equal(r.shape, intens.shape))
        
        # funcs and params
        assert(len(funcs)>=1)
        for i in range(len(funcs)):
            assert(funcs[i] in ncempy.algo.math.lkp_funcs)
        
        init_guess = np.array(init_guess)
        init_guess = np.reshape(init_guess, sum(map(lambda x: ncempy.algo.math.lkp_funcs[x][1], funcs)))

    except:
        raise TypeError('Something wrong with the input!')
    
    if maxfev is None:
        maxfev = 1000
 
    popt, flag = scipy.optimize.leastsq( residuals_fit, init_guess, args=(r, intens, funcs), maxfev=maxfev)

    if flag not in [1,2,3,4]:
        print('WARNING: fitting of radial profile failed.')

    return popt


def plot_fit( r, intens, dims, funcs, param, show=False ):
    '''Plot the fit results to the radial profile.
    
    Parameters:
        r (np.ndarray):    r-axis of radial profile.
        intens (np.ndarray):    Intensity-axis of radial profile.
        dims (tuple):    Dimensions of original image to read out units.
        funcs (tuple):    List of functions.
        param (np.ndarray):    Parameters for functions in funcs.
        show (bool):    Set to directly show plot interactively.
    
    Returns:
        (np.ndarray):    Image of the plot.
        
    '''
    
    try:
        # check data
        assert(isinstance(r, np.ndarray))
        assert(isinstance(intens, np.ndarray))
        assert(np.array_equal(r.shape, intens.shape))
        
        # check if dims availabel
        assert(len(dims)>=1)
        assert(len(dims[0])==3)
        
        # funcs and params
        assert(len(funcs)>=1)
        for i in range(len(funcs)):
            assert(funcs[i] in ncempy.algo.math.lkp_funcs)
            
        param = np.array(param)
        param = np.reshape(param, sum(map(lambda x: ncempy.algo.math.lkp_funcs[x][1], funcs)))

    except:
        raise TypeError('Something wrong with the input!')
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    # plot radial profile
    ax.plot(r, intens, 'r-')
    
    # plot single
    n = 0
    for i in range(len(funcs)):
        ax.plot(r, ncempy.algo.math.lkp_funcs[funcs[i]][0]( r, param[n:n+ncempy.algo.math.lkp_funcs[funcs[i]][1]]), 'g-' )
        n += ncempy.algo.math.lkp_funcs[funcs[i]][1]
    # sum of functions
    sum_funcs = ncempy.algo.math.sum_functions( r, funcs, param )
    ax.plot(r, sum_funcs, 'b-')
    
    # labels
    ax.set_xlabel('r /{}'.format(dims[0][2]))
    ax.set_ylabel('I /[a.u.]')
        
    if show:
        plt.show(block=False)
    
    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    return plot
    
    
def run_singleImage( img, dims, settings, show=False ):
    '''Evaluate a single ring diffraction pattern with given settings.
    
    Parameters:
        img (np.ndarray):    Image.
        dims (tuple):    Corresponding dim vectors.
        settings (dict):    Dict of settings necessary for the evaluation.
        show (bool):    Set to directly show plots interactively.

    Returns:
        (np.ndarray):    Optimized parameters of fitting the radial profile according to the settings.
        
    '''
    
    try:
        assert(isinstance(img, np.ndarray))
        
        # check if dims availabel
        assert(len(dims)>=1)
        assert(len(dims[0])==3)
        
        assert(type(settings) is dict)
        
    except:
        raise RuntimeError('Something wrong with the input')
        
    
    # create a copy of settings to decouple
    mysettings = copy.deepcopy(settings)
    
    # get local maxima an turn them into real space coords
    points = ncempy.algo.local_max.local_max(img, mysettings['lmax_r'], mysettings['lmax_thresh'])
    points = ncempy.algo.local_max.points_todim(points, dims)
    
    # convert center to real space
    center_init = ncempy.algo.local_max.points_todim(mysettings['lmax_cinit'], dims)
    
    # filter to single ring
    points = ncempy.algo.distortion.filter_ring(points, center_init, mysettings['lmax_range'])
    
    if show:
        if settings['plt_imgminmax'] is None:
            mysettings['plt_imgminmax'] = (0.,1.)
            
        plot = ncempy.algo.local_max.plot_points(img, points, vminmax=mysettings['plt_imgminmax'], dims=dims, invert=True, show=show)
    
    # optimize center
    center = ncempy.algo.distortion.optimize_center(points, center_init, verbose=show)
        
    # fit distortions
    points_plr = ncempy.algo.distortion.points_topolar(points, center)
    dists = ncempy.algo.distortion.optimize_distortion(points_plr, mysettings['ns'])
    if show:
        plot = ncempy.algo.distortion.plot_distpolar(points_plr, dims, dists, mysettings['ns'], show=show)
    
    # calc coordinates in optimized system
    rs, thes = ncempy.algo.radial_profile.calc_polarcoords( center, dims, mysettings['ns'], dists )
    
    if settings['rad_rmax'] is None:
        mysettings['rad_rmax'] = np.abs(dims[0][0][0]-dims[0][0][1])*np.min(img.shape)/2.0
    if settings['rad_dr'] is None:
        mysettings['rad_dr'] = np.abs(dims[0][0][0]-dims[0][0][1])/10.
    if settings['rad_sigma'] is None:
        mysettings['rad_sigma'] = np.abs(dims[0][0][0]-dims[0][0][1])
        
    # extract radial profile
    R, I = ncempy.algo.radial_profile.calc_radialprofile( img, rs, mysettings['rad_rmax'], mysettings['rad_dr'], mysettings['rad_sigma'], mask=mysettings['mask'] )
    
    # cut radial profile
    sel = (R>=mysettings['fit_rrange'][0])*(R<=mysettings['fit_rrange'][1])
    I = I[sel]
    R = R[sel]
    
    rawRI = np.copy(np.array([R,I]).transpose())
    
    # subtract a power law background fitted to specific points
    fit_R = np.array([])
    fit_I = np.array([])
    for xpoint in mysettings['back_xs']:
        ix = np.where( np.abs(R-xpoint) < mysettings['back_xswidth'] )
        fit_R = np.append(fit_R, R[ix])
        fit_I = np.append(fit_I, I[ix])
            
    funcs_back = [ 'const', 'powlaw']
    res_back = ncempy.algo.radial_profile.fit_radialprofile( fit_R, fit_I, funcs_back, mysettings['back_init'], maxfev=1000 )
    if show:
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs_back, res_back, show=show )
        
    I = I - ncempy.algo.math.sum_functions(R, funcs_back, res_back)    
    
    
    # fit
    res = ncempy.algo.radial_profile.fit_radialprofile( R, I, mysettings['fit_funcs'], mysettings['fit_init'], maxfev=mysettings['fit_maxfev'])
    
    if show:
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, mysettings['fit_funcs'], res, show=show )
        
    return np.array([R,I]).transpose(), res, center, dists, rawRI, res_back, mysettings
    



'''
Module to process ring diffraction patterns.
'''

import ncempy.algo.local_max
import ncempy.algo.distortion
import ncempy.algo.radial_profile
import ncempy.algo.math
import ncempy.io.emd

import matplotlib.pyplot as plt
import numpy as np
import h5py


# used to indicate settings format
cur_set_vers = 'ring_diffraction_setting_vers0.1'
'''(str):    Used to identify settings hdf5 groups.'''

cur_eva_vers = 'ring_diffraction_evaluation_vers0.1'
'''(str):    Used to identify evaluation hdf5 groups.'''

dummie_settings = { 'lmax_r': 1,
                    'lmax_thresh': 1,
                    'lmax_cinit': (1,1),
                    'lmax_range': (1.,2.),
                    'plt_imgminmax': (0.0,1.0),
                    'ns': (1,),
                    'rad_rmax': 1,
                    'rad_dr': 0.1,
                    'rad_sigma': 0.1,
                    'mask': np.ones((5,5)),
                    'fit_rrange': (1.0, 2.0),
                    'back_xs': (1.0, 2.0, 3.0),
                    'back_xswidth': 0.1,
                    'back_init': (1.0, 1.0, 1.0),
                    'fit_funcs': ('voigt',),
                    'fit_init': (1.0, 1.0, 1.0, 1.0),
                    'fit_maxfev':10
                  }
'''(dict):    Dummy settings with all parameters set.'''

min_dummie_settings = { 'lmax_r': 1,
                        'lmax_thresh': 1,
                        'lmax_cinit': (1,1),
                        'lmax_range': (1.,2.),
                        'plt_imgminmax': None,
                        'ns': (1,),
                        'rad_rmax': None,
                        'rad_dr': None,
                        'rad_sigma': None,
                        'mask': None,
                        'fit_rrange': (1.0, 2.0),
                        'back_xs': (1.0, 2.0, 3.0),
                        'back_xswidth': 0.1,
                        'back_init': (1.0, 1.0, 1.0),
                        'fit_funcs': ('voigt',),
                        'fit_init': (1.0, 1.0, 1.0, 1.0),
                        'fit_maxfev': None
                    }
'''(dict):    Dummy settings with all parameters set but all optional ones as Nones.'''


def get_settings( parent ):
    '''Get settings for radial profile evaluation.
    
    Parameters:
        parent (h5py._hl.group.Group):    Input group.
    
    Returns:
        (dict):    Settings read from parent.
        
    '''
    
    try:
        assert( isinstance(parent, h5py._hl.group.Group) )
    except:
        raise TypeError('Something wrong with the input.')
    
    
    if not parent.attrs['type'] == np.string_(cur_set_vers):
        print('Don\'t  know the format of these settings.')
        return None
    
    settings = {}
    
    settings['lmax_r'] = parent.attrs['lmax_r']
    settings['lmax_thresh'] = parent.attrs['lmax_thresh']
    settings['lmax_cinit'] = parent.attrs['lmax_cinit']
    settings['lmax_range'] = parent.attrs['lmax_range']
    settings['ns'] = parent.attrs['ns']
    settings['fit_rrange'] = parent.attrs['fit_rrange']
    settings['back_xs'] = parent.attrs['back_xs']
    settings['back_xswidth'] = parent.attrs['back_xswidth']
    settings['back_init'] = parent.attrs['back_init']
    settings['fit_init'] = parent.attrs['fit_init']
    
    
    if isinstance(parent.attrs['fit_funcs'], np.ndarray):
        in_funcs = parent.attrs['fit_funcs'][0]
    else:
        in_funcs = parent.attrs['fit_funcs']
    in_funcs = in_funcs.split(b',')
        
    out_funcs = []
    for i in range(len(in_funcs)):
        out_funcs.append(in_funcs[i].decode('utf-8').strip())
    settings['fit_funcs'] = tuple(out_funcs)
    
    
    if 'plt_imgminmax' in parent.attrs:
        settings['plt_imgminmax'] = parent.attrs['plt_imgminmax']
    else:
        settings['plt_imgminmax'] = None

    if 'rad_rmax' in parent.attrs:
        settings['rad_rmax'] = parent.attrs['rad_rmax']
    else:
        settings['rad_rmax'] = None

    if 'rad_dr' in parent.attrs:
        settings['rad_dr'] = parent.attrs['rad_dr']
    else:
        settings['rad_dr'] = None

    if 'rad_sigma' in parent.attrs:
        settings['rad_sigma'] = parent.attrs['rad_sigma']
    else:
        settings['rad_sigma'] = None

    if 'mask' in parent:
        settings['mask'] = np.copy(parent['mask'])
    else:
        settings['mask'] = None

    if 'fit_maxfev' in parent.attrs:
        settings['fit_maxfev'] = parent.attrs['fit_maxfev']
    else:
        settings['fit_maxfev'] = None

    return settings


def put_settings( parent, settings ):
    '''Put settings for radial profile evaluation.
    
    Creates a subgroup in parent holding the settings as attributes.
    
    Parameters:
        parent (h5py._hl.group.Group):    Group to hold settings subgroup.
        setting (dict):    Settings to write.
    
    Returns:
        (h5py._hl.group.Group):    Handle to settings group.
        
    '''
 
    try:
        assert( isinstance(parent, h5py._hl.group.Group) )
        assert( type(settings) is dict )
    except:
        raise TypeError('Something wrong with the input.')
    
  
    try:
        grp_set = parent.create_group('settings_ringdiffraction')
    except:
        raise RuntimeError('Could not write to {}'.format(parent))
        
        
    # set version information    
    grp_set.attrs['type'] = np.string_(cur_set_vers)
    
    # hardcoding the written settings to keep control
    grp_set.attrs['lmax_r'] = settings['lmax_r']
    grp_set.attrs['lmax_thresh'] = settings['lmax_thresh']
    grp_set.attrs['lmax_cinit'] = settings['lmax_cinit']
    grp_set.attrs['lmax_range'] = settings['lmax_range']
    grp_set.attrs['ns'] = settings['ns']
    grp_set.attrs['fit_rrange'] = settings['fit_rrange']
    grp_set.attrs['back_xs'] = settings['back_xs']
    grp_set.attrs['back_xswidth'] = settings['back_xswidth']
    grp_set.attrs['back_init'] = settings['back_init']
    grp_set.attrs['fit_init'] = settings['fit_init']
    
    fit_funcs = []
    for i in range(len(settings['fit_funcs'])):
        fit_funcs.append(settings['fit_funcs'][i])
    grp_set.attrs['fit_funcs'] = np.string_(', '.join(fit_funcs))
    
    if not settings['plt_imgminmax'] is None:
        grp_set.attrs['plt_imgminmax'] = settings['plt_imgminmax']
    if not settings['rad_rmax'] is None:
        grp_set.attrs['rad_rmax'] = settings['rad_rmax']
    if not settings['rad_dr'] is None:
        grp_set.attrs['rad_dr'] = settings['rad_dr']
    if not settings['rad_sigma'] is None:
        grp_set.attrs['rad_sigma'] = settings['rad_sigma']
    if not settings['mask'] is None:
        grp_set.create_dataset('mask', data=settings['mask'])
    if not settings['fit_maxfev'] is None:
        grp_set.attrs['fit_maxfev'] = settings['fit_maxfev']
        
    return grp_set


def put_sglgroup(parent, label, data_grp):
    '''Puts the todo evaluation into parent.
    
    Remember that the ressource of the external link must not be already opened elsewhere to access data.
    
    Parameters:
        parent (h5py._hl.group.Group):    Hdf5 group to add this evaluation group to.
        label (str):    Label for the evaluation group.
        data_grp (h5py._hl.group.Group):    Emdtype group where to find the data.
    
    Returns:
        (h5py._hl.group.Group):    Handle to group.
        
    '''
    
    try:
        assert( isinstance(parent, h5py._hl.group.Group) )
        label = str(label)
        assert( isinstance(data_grp, h5py._hl.group.Group) )
    except:
        raise TypeError('Something wrong with the input')
        
    
    # create the evaluation group
    grp = parent.create_group(label)
    grp.attrs['type'] = np.string_(cur_eva_vers)
    
    # put a link to the data
    grp.attrs['filename'] = np.string_(data_grp.file.filename)
    grp.attrs['internal_path'] = np.string_(data_grp.name)
    #grp['emdgroup'] = h5py.ExternalLink(data_grp.file.filename, data_grp.name)    
    
    return grp
    

def run_sglgroup(group, outfile, overwrite=False, verbose=False, showplots=False):
    '''Run evaluation on a single group.
    
    Parameters:
        group (h5py._hl.group.Group):    Handle to evaluation group to execute.
        outfile (ncempy.io.emd.fileEMD):    Emdfile for output.
        overwrite (bool):    Set to overwrite existing results in outfile.
        verbose (bool):    Set to get verbose output during run.
        showplots (bool):    Set to directly show plots interactively.
        
    '''

    try:
        assert(isinstance(group, h5py._hl.group.Group))
        assert( group.attrs['type'] == np.string_(cur_eva_vers) )
        
        assert(isinstance(outfile, ncempy.io.emd.fileEMD))
    except:
        raise TypeError('Something wrong with the input.')
        
    
    if verbose:
        print('Running evaluation of "{}".'.format(group.name))
        
        
    # get the emdgroup
    if verbose:
        print('.. getting data from {}:{}'.format(group.attrs['filename'].decode('utf-8'), group.attrs['internal_path'].decode('utf-8')))
    readfile = ncempy.io.emd.fileEMD( group.attrs['filename'].decode('utf-8'), readonly=True )
    data, dims = readfile.get_emdgroup(readfile.file_hdl[group.attrs['internal_path'].decode('utf-8')])
    
    # find the settings moving upwards in hierarchy
    if verbose:
        print('.. searching for settings.')
    def proc_group(grp):
        #print('scanning group {}'.format(grp))
        if 'settings_ringdiffraction' in grp:
            stt = grp['settings_ringdiffraction']
            if stt.attrs['type'] == np.string_(cur_set_vers):
                return stt
        else:
            if not grp == grp.file:
                return proc_group(grp.parent)

    grp_set = proc_group(group)
    
    if grp_set is None:
        raise RuntimeError('Could not find settings in evaluation group or its parents.')
    else:
        if verbose:
            print('.. loading settings from {}.'.format(grp_set.name))
        settings = get_settings(grp_set)
    
    # what data to collect
    profiles = None
    centers = None
    distss = None
    fits = None
    rawprofiles = None
    fits_back = None
    
    # run evaluation with settings
    if len(data.shape) == 3:
        for i in range(data.shape[0]):
            profile, res, center, dists, rawprofile, res_back, myset = ncempy.algo.radial_profile.run_singleImage( data[i,:,:], dims[1:3], settings,  show=showplots)
    
            # after first run I know the size
            if profiles is None:
                profiles = np.zeros( (profile.shape[0], data.shape[0]) )
                fits = np.zeros( (res.shape[0], data.shape[0]) )
                centers = np.zeros( (2, data.shape[0]) )
                distss = np.zeros( (dists.shape[0], data.shape[0]) )
                rawprofiles = np.zeros( (rawprofile.shape[0], data.shape[0]) )
                fits_back = np.zeros( (res_back.shape[0], data.shape[0]) )
                
            # assign data    
            profiles[:,i] = profile[:,1]
            fits[:,i] = res[:]
            centers[:,i] = center[:]
            distss[:,i] = dists[:]
            rawprofiles[:,i] = rawprofile[:,1]
            fits_back[:,i] = res_back[:]

        # save results in this group
        outfile.put_emdgroup('radial_profile', profiles, ( (profile[:,0], 'radial distance', dims[2][2]) , dims[0]), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('fit_results', fits, ( ( np.array(range(fits.shape[0])), 'parameters', '[]') , dims[0]), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('centers', centers, ( ( np.array(range(2)), 'dimension', dims[2][2]) , dims[0]), parent=group, overwrite=overwrite)    
        outfile.put_emdgroup('distortions', distss, ( ( np.array(range(distss.shape[0])), 'parameters', '[]') , dims[0]), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('radial_profile_noback', rawprofiles, ( (rawprofile[:,0], 'radial distance', dims[2][2]) , dims[0]), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('back_results', fits_back, ( ( np.array(range(fits_back.shape[0])), 'background parameters', '[]') , dims[0]), parent=group, overwrite=overwrite)
    

    elif len(data.shape) == 2:
        profile, res, center, dists, rawprofile, res_back, myset = ncempy.algo.radial_profile.run_singleImage( data[:,:], dims, settings,  show=showplots)
        
        # save results in this group
        outfile.put_emdgroup('radial_profile', profile[:,1], ( (profile[:,0], 'radial distance', dims[0][2]), ), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('fit_results', res, ( ( np.array(range(res.shape[0])), 'parameters', '[]'), ), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('centers', center, ( ( np.array(range(2)), 'dimension', dims[0][2]), ), parent=group, overwrite=overwrite)    
        outfile.put_emdgroup('distortions', dists, ( ( np.array(range(dists.shape[0])), 'parameters', '[]'), ), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('radial_profile_noback', rawprofile[:,1], ( (rawprofile[:,0], 'radial distance', dims[0][2]), ), parent=group, overwrite=overwrite)
        outfile.put_emdgroup('back_results', res_back, ( ( np.array(range(res_back.shape[0])), 'background parameters', '[]'), ), parent=group, overwrite=overwrite)

    else:
        raise RuntimeError('Cannot handle that data.')

    # save a log comment
    outfile.put_comment('Evaluated "{}" using ring diffraction analysis.'.format(group.name))
    

def run_all(parent, outfile, overwrite=False, verbose=False, showplots=False):
    '''
    Run on a set-up emd file to do evaluations and save results.
    
    All evaluations within parent are run.
    
    Parameters:
        parent (h5py._hl.group.Group):    Handle to parent.
        outfile (ncempy.io.emd.fileEMD):    Emdfile for output.
        overwrite (bool):    Set to overwrite existing results in outfile.
        verbose (bool):    Set to get verbose output during run.
        showplots (bool):    Set to directly show plots interactively.
        
    '''
    
    # get all groups with evaluations to do
    todo = []
    
    # recursive function
    def proc_group(grp, todo):
        for item in grp:
            if grp.get(item, getclass=True) == h5py._hl.group.Group:
                item = grp.get(item)
                if 'type' in item.attrs:
                    if item.attrs['type'] == np.string_(cur_eva_vers):
                        todo.append(item)
                        if verbose:
                            print('Found evaluation group at "{}".'.format(item.name) )
                proc_group(item, todo)    
    
    proc_group(parent, todo)
    
    # run through all evaluations
    for i in range(len(todo)):
        run_sglgroup(todo[i], outfile, overwrite=overwrite, verbose=verbose, showplots=showplots)
    

